#                 ======================================
#		           R library  "elliptical"
#                 ====================================== 
#
#               Copyright (C) 2005 Francisco Jos A. Cysneiros
#
#
#  This library is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published
#  by the Free Software Foundation; either version 2 of the License,
#  or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  
#  USA or look up the web page http://www.gnu.org/copyleft/gpl.html.
#
#  Please send any comments, suggestions or errors found to:
#  Francisco Jos A. Cysneiros, Departamento de Estatstica - CCEN,
#  Universidade Federal de Pernambuco 
#  Recife (PE), Brazil.  Email: cysneiros@de.ufpe.br. 
#  Web: http://www.de.ufpe.br/~cysneiros.
# ==============================================================================
#
#                         User-defined "elliptical" families
#                                             
# ==============================================================================
# NOTE: This functions will be used in the demonstration file associated 
#        with the "elliptical" section of library "elliptical". 
# This example is consist in modelling the stack date for linear regression 
# This another example is consist in modelling the  density date  for nonlinear regression


library(Matrix)
####example
stack.dat<-scan(what=list(air=0,water=0,acid=0,ammonia=0))
80 27 89 42
80 27 88 37
75 25 90 37
62 24 87 28
62 22 87 18
62 23 87 18
62 24 93 19
62 24 93 20
58 23 87 15
58 18 80 14
58 18 89 14
58 17 88 13
58 18 82 11
58 19 93 12
50 18 89  8
50 18 86  7
50 19 72  8
50 19 79  8
50 20 80  9
56 20 82 15
70 20 91 15

attach(stack.dat)
ammonia<-stack.dat$ammonia
air<-stack.dat$air
water<-stack.dat$water
acid<-stack.dat$acid
#model ammonia=alpha*intercept +beta*air + nu*water + theta acid
stack<-data.frame(ammonia,air,water,acid)

stack.elpt<-elliptical(formula=ammonia~air+water+acid,family=Student(4),data=stack)
summary(stack.elpt)
##fit under x1=0
Cres<-matrix(c(0,1,0,0),1,4)
sol<-matrix(0,1,1)
stackrest.elpt<-elliptical(formula=ammonia~air+water+acid,family=Student(4),data=stack,restrict=T,Cres=Cres,sol=sol)
summary(stackrest.elpt)
##diagnostic's analysis
stackdiag<-elliptical.diag(stack.elpt)# calculte diagnotics measures
 arg<-4
 envelope(stack.elpt,arg=arg)
### if you want identify points use iden=T , if no, don't use.
elliptical.diag.plots(stack.elpt,iden=T) # plot diagnotics
					  

#Model  3.2 Holliday log(Y)=(alpha+betaX+gammaX^2)^{-1}
# Ratkowsky, D.A (1983) Nonlinear regression modeling section 3.1
#date in appendix 3.A pg 58
onion.dat<-scan(what=list(x=0,y=0))
22.30 148.57
25.86 125.30
29.09 150.69
29.74 147.42
31.68 117.10
31.68 116.64
32.00 129.66
32.32 131.54
32.32 151.50
34.91 121.80
35.23 125.67
38.47 117.78
39.44 101.50
41.05 113.22
41.70 136.43
44.28 117.54
45.90  87.20
46.55 107.41
48.16 129.68
48.49 104.63
48.81 114.15
49.78  99.85
50.43 111.65
51.72  98.09
61.42  87.85
65.29  75.45
67.23  87.01
71.44  90.10
73.05  81.08
86.63  65.33
96.00  58.49
98.91  65.67
103.44 67.19
105.05 54.01
111.19 60.92
113.78 53.48
119.92 61.62
120.89 26.32
126.71 61.21
138.99 41.67
146.75 45.26
160.97 46.45

attach(onion.dat)
y<-log(onion.dat$y)
x1<-onion.dat$x
x2<-onion.dat$x^2
X <- cbind(1,x1,x2)
#model y= -log(alpha + beta*x1+gamma*x2)
DerBonion<- function(parm,X)
{		
         alpha<-parm[1]
         beta<-parm[2]
		 gamma<-parm[3]
		 x0<-X[,1]
		 x1<-X[,2]
		 x2<-X[,3]
		 den<-alpha*x0+beta*x1+gamma*x2
	    .grad <- array(0, c(length(x0), 3), list(NULL, c("alpha","beta","gamma")))
		.grad[, "alpha"]<- -x0/den
		.grad[, "beta"] <- -x1/den
		.grad[, "gamma"]   <- -x2/den
		.value<--log(den)
		.hess <- array(0, c(3, 3,length(x0)), list( c("alpha","beta","gamma"), c("alpha","beta","gamma"),NULL))
		#necessary, if you want to do diagnostic
		.hess["alpha","alpha",]<-(x0^2)*(den)^(-2)
		.hess["alpha","beta",]<-(x0*x1)*(den)^(-2)
        .hess["alpha","gamma",]<-(x0*x2)*(den)^(-2)
		.hess["beta","alpha",]<-(x1*x0)*(den)^(-2)
		.hess["beta","beta",]<-(x1^2)*(den)^(-2)
		.hess["beta","gamma",]<-(x1*x2)*(den)^(-2)
        .hess["gamma","alpha",]<-(x2*x0)*(den)^(-2)
		.hess["gamma","beta",]<-(x1*x2)*(den)^(-2)
        .hess["gamma","gamma",]<-(x2^2)*(den)^(-2)
		fit <- list(value=.value,gradient=.grad,hessian=.hess)		
	    fit
}
onion<-data.frame(y,x1,x2)
onion.nonelpt<-elliptical(formula=y~x1+x2,linear=F,DerB=DerBonion,parmB=c(0.001,0.1e-3,0.1e-6),family=LogisII(),data=onion)
summary(onion.nonelpt)
win.graph()
par(mfrow=c(2,2))
envelope(onion.nonelpt,DerB=DerBonion)
elliptical.diag.plots(onion.nonelpt,iden=T)
### if you want identify points use iden=T , if no, don't use.
